#!

;;;============================================================================

;;; File: "stoe.scm", Time-stamp: <2007-06-04 23:15:28 feeley>

;;; Copyright (c) 2007 by Marc Feeley, All Rights Reserved.

;;; A prototype Scheme to Erlang compiler based on the Gambit-C compiler.

;;; The compiler does not handle all of Scheme.  Here are the main
;;; restrictions.
;;;
;;;   - letrec is not supported (Erlang only allows mutually recursive
;;;     functions at top level)
;;;
;;;   - set! of global variables is not supported (Erlang is side-effect
;;;     free but cells for local variables can be implemented using
;;;     processes)
;;;
;;;   - eq? is not supported (object identity is a concept associated
;;;     with Erlang "references" only)
;;;
;;;   - only fixed arity lambda expressions and function definitions
;;;     are supported
;;;
;;;   - the Scheme file must only contain function definitions and
;;;     all the functions are exported by the Erlang module that
;;;     is generated
;;;
;;;   - the mapping to Erlang is as direct as possible, so some
;;;     details of the Scheme semantics are not properly implemented
;;;     (for example both the Scheme value #f and the symbol false are
;;;     mapped to the Erlang atom 'false', the Scheme character #\!
;;;     is mapped to the Erlang integer 33, etc)
;;;
;;;   - a few Scheme primitives are implemented (and also the ``!'',
;;;     ``?'' and ``self'' Termite primitives)

;;; Sample usage:
;;;
;;; % cat test.scm
;;; (define (fib n)
;;;   (if (< n 2) 1 (+ (fib (- n 1)) (fib (- n 2)))))
;;; 
;;; (define (fact n)
;;;   (if (< n 2) 1 (* n (fact (- n 1)))))
;;; 
;;; (define (iota n)
;;;   (iota-2 n '()))
;;; 
;;; (define (iota-2 i lst)
;;;   (if (> i 0)
;;;       (iota-2 (- i 1) (cons i lst))
;;;       lst))
;;; 
;;; (define (tally)
;;;   (let ((sum 0))
;;;     (lambda (x)
;;;       (set! sum (+ sum x))
;;;       sum)))
;;; 
;;; (define (process n dest)
;;;   (lambda () (! dest (map (?) (iota n)))))
;;; 
;;; (define (main)
;;;   (let ((a (spawn (process 10 (self))))
;;;         (b (spawn (process 10 (self))))
;;;         (t1 (tally))
;;;         (t2 (tally)))
;;;     (! a fib)
;;;     (! b fact)
;;;     (map t1 (?))
;;;     (map t2 (?))
;;;     (list (t1 0) (t2 0))))
;;; % gsc -i stoe.scm test.scm
;;; % cat test.erl
;;; -module(test).
;;; -export([fib/1,fact/1,iota/1,iota@2d@2/2,tally/0,process/2,main/0]).
;;; 
;;; fib(N) ->
;;;   if (N < 2) -> 1; true -> (fib((N - 1)) + fib((N - 2))) end.
;;; 
;;; fact(N) ->
;;;   if (N < 2) -> 1; true -> (N * fact((N - 1))) end.
;;; 
;;; iota(N) ->
;;;   iota@2d@2(N,[]).
;;; 
;;; iota@2d@2(I,Lst) ->
;;;   if (I > 0) -> iota@2d@2((I - 1),[I|Lst]); true -> Lst end.
;;; 
;;; tally() ->
;;;   Sum = schemelib:box(0),
;;;   fun (X) -> schemelib:setbox(Sum,(schemelib:unbox(Sum) + X)),
;;;     schemelib:unbox(Sum) end.
;;; 
;;; process(N,Dest) ->
;;;   fun () -> (Dest ! lists:map(receive Tmp_1 -> Tmp_1 end,iota(N))) end.
;;; 
;;; main() ->
;;;   {T2,T1,B,A} = {tally(),tally(),spawn(process(10,self())),spawn(process(10,self()))},
;;;   (A ! fun fib/1),
;;;   (B ! fun fact/1),
;;;   lists:map(T1,receive Tmp_2 -> Tmp_2 end),
;;;   lists:map(T2,receive Tmp_3 -> Tmp_3 end),
;;;   [T1(0),T2(0)].
;;; 
;;; % erl
;;; Eshell V5.5.4  (abort with ^G)
;;; 1> c(schemelib.erl).
;;; {ok,schemelib}
;;; 2> c(test.erl).
;;; {ok,test}
;;; 3> test:main().
;;; [4037913,231]

;;;============================================================================

(##namespace ("c#")) ;; use the Gambit compiler's namespace
(##include "~~/lib/gambit#.scm") ;; except for the Gambit primitives

(define module-name #f)
(define local-funs #f)

(define (s2e input)

  (define info-port #f)
  (define target-name 'c)

  (set! module-name (path-strip-directory (path-strip-extension input)))

  (env.begin!)
  (ptree.begin! info-port)
  (virtual.begin!)
  (target-select! target-name info-port)

  (let ((x (read-source input #f #t)))
    (parse-program
     (expand-source (wrap-program (##vector-ref x 1)))
     (make-global-environment)
     module-name
     (lambda (lst env c-intf)
       (let ((lst (lambda-lift (beta-reduce (assignment-convert lst)))))
         (generate-module lst)))))

  (target-unselect!)
  (virtual.end!)
  (ptree.end!)
  (env.end!))

(define (generate-module lst)
  (let ((defs (keep def? lst))
        (exps (keep (lambda (x) (not (def? x))) lst)))
    (if (not (null? exps))
        (error "Scheme file must only contain definitions")
        (with-output-to-file
            (string-append module-name ".erl")
          (lambda ()
            (set! local-funs
                  (map
                   (lambda (def)
                     (let ((var (def-var def))
                           (val (def-val def)))
                       (if (prc? val)
                           (let ((parms (prc-parms val)))
                             (cons (var-name var)
                                   (length parms)))
                           (error "Only function definitions are allowed"))))
                   defs))
            (generate-module-header)
            (generate-module-body defs))))))

(define (generate-module-header)
  (display
   (list "-module("
         module-name
         ").\n"
         "-export("
         (pe-list local-funs "[" "]" fun-and-arity)
         ").\n\n")))

(define (generate-module-body defs)
  (for-each generate-definition defs))

(define (generate-definition ptree)
  (let ((var (def-var ptree))
        (val (def-val ptree)))
    (if (prc? val)
        (let* ((ctx (make-top-ctx))
               (parms (prc-parms val))
               (parm-names (pe-add-vars parms ctx)))
          (display
           (list (pe-fun-name (var-name var))
                 (pe-list parm-names "(" ")" (lambda (x) x))
                 " ->\n"
                 (indent 1)
                 (pe (prc-body val) ctx)
                 ".\n\n")))
        (error "Only function definitions are allowed"))))

(define (make-top-ctx)
  (vector 1 #t (vector '())))

(define (ctx-indent ctx)
  (vector-ref ctx 0))

(define (ctx-indent-set ctx indent)
  (vector indent (vector-ref ctx 1) (vector-ref ctx 2)))

(define (ctx-body? ctx)
  (vector-ref ctx 1))

(define (ctx-body?-set ctx body?)
  (vector (vector-ref ctx 0) body? (vector-ref ctx 2)))

(define (ctx-lvars ctx)
  (vector-ref ctx 2))

(define (ctx-lvars-set ctx lvars)
  (vector (vector-ref ctx 0) (vector-ref ctx 1) lvars))

(define (fun-and-arity x)
  (let ((name (car x))
        (arity (cdr x)))
    (string-append (pe-fun-name name) "/" (number->string arity))))

(define (pe-function name)
  (let ((x (assq name local-funs)))
    (if x
        (string-append "fun " (fun-and-arity x))
        (error "unknown function:" name))))

(define (pe-fun-name name)
  (apply
   string-append
   (map (lambda (c)
          (if (or (char-alphabetic? c) (char-numeric? c))
              (string c)
              (let ((n (char->integer c)))
                (define (hex i) (string-ref "0123456789abcdef" i))
                (string #\@ (hex (quotient n 16)) (hex (modulo n 16)) #\@))))
        (string->list (symbol->string name)))))

(define (pe-variable var ctx)
  (let ((lvars (ctx-lvars ctx)))
    (let ((x (assq var (vector-ref lvars 0))))
      (if x
          (cdr x)
          #f))))

(define (pe-add-vars vars ctx)
  (let ((lvars (ctx-lvars ctx)))
    (map (lambda (v)
           (let ((x (assq v (vector-ref lvars 0))))
             (if x
                 (cdr x)
                 (let ((name0
                        (let ((lst
                               (string->list (symbol->string (var-name v)))))
                          (list->string
                           (keep (lambda (c)
                                   (or (char-alphabetic? c)
                                       (char-numeric? c)))
                                 (cons (char-upcase (car lst)) (cdr lst)))))))
                   (let loop1 ((i 2) (name name0))
                     (let loop2 ((lst (vector-ref lvars 0)))
                       (if (pair? lst)
                           (if (string=? (cdr (car lst)) name)
                               (loop1 (+ i 1)
                                      (string-append name0
                                                     "_"
                                                     (number->string i)))
                               (loop2 (cdr lst)))
                           (begin
                             (vector-set! lvars 0
                               (cons (cons v name) (vector-ref lvars 0)))
                             name))))))))
         vars)))

(define newtmp
  (let ((n 0))
    (lambda ()
      (set! n (+ n 1))
      (string->symbol (string-append "Tmp_" (number->string n))))))

(define (parse-tree->erlang ptree)
  (pe ptree (make-top-ctx)))

(define (pe ptree ctx)

  (define (pe-plain ptree ctx)
    (pe ptree (ctx-body?-set ctx #f)))

  (define (pe-body ptree ctx)
    (pe ptree (ctx-indent-set (ctx-body?-set ctx #t) (+ 1 (ctx-indent ctx)))))

  (cond ((cst? ptree)
         (let ((val (cst-val ptree)))
           (pe-constant val)))

        ((ref? ptree)
         (let ((var (ref-var ptree)))
           (or (pe-variable var ctx)
               (pe-function (var-name var)))))

        ((set? ptree)
         (error "set! of global variables are not allowed"))

        ((def? ptree)
         (error "internal error")) ;; should never happen

        ((tst? ptree)
         (list "if "
               (pe-plain (tst-pre ptree) ctx)
               " -> "
               (pe-body (tst-con ptree) ctx)
               "; true -> "
               (pe-body (tst-alt ptree) ctx)
               " end"))

        ((conj? ptree)
         (list "("
               (pe-plain (conj-pre ptree) ctx)
               " andalso "
               (pe-plain (conj-alt ptree) ctx)
               ")"))

        ((disj? ptree)
         (list "("
               (pe-plain (disj-pre ptree) ctx)
               " orelse "
               (pe-plain (disj-alt ptree) ctx)
               ")"))

        ((prc? ptree)
         (let* ((parms (prc-parms ptree))
                (parm-names (pe-add-vars parms ctx)))
           (list "fun "
                 (pe-list parm-names "(" ")" (lambda (x) x))
                 " -> "
                 (pe-body (prc-body ptree) (ctx-body?-set ctx #f))
                 " end")))

        ((app? ptree)
         (let ((oper (app-oper ptree))
               (args (app-args ptree)))
           (if (and (prc? oper) ; applying a lambda-expr is like a 'let'
                    (prc-req-and-opt-parms-only? oper)
                    (= (length (prc-parms oper)) (length args)))
               (if (varset-intersects?
                    (list->varset (prc-parms oper))
                    (varset-union-multi (map bound-free-variables args)))
                   (error "letrec is not supported")
                   (let* ((parms (prc-parms oper))
                          (parm-names (pe-add-vars parms ctx)))

                     (define (bindings ctx)
                       (list
                        (cond ((null? parms)
                               "")
                              ((null? (cdr parms))
                               (if (var-useless? (car parms))
                                   (list (pe-plain (car args) ctx)
                                         ",\n"
                                         (indent (ctx-indent ctx)))
                                   (list (car parm-names)
                                         " = "
                                         (pe-plain (car args) ctx)
                                         ",\n"
                                         (indent (ctx-indent ctx)))))
                              (else
                               (list (pe-list parm-names "{" "}"
                                              (lambda (x) x))
                                     " = "
                                     (pe-list args "{" "}"
                                              (lambda (x) (pe-plain x ctx)))
                                     ",\n"
                                     (indent (ctx-indent ctx)))))))

                     (if (ctx-body? ctx)
                         (list
                          (bindings ctx)
                          (pe (prc-body oper) ctx))
                         (list
                          "begin "
                          (bindings (ctx-indent-set ctx (+ 1 (ctx-indent ctx))))
                          (pe-body (prc-body oper) ctx)
                          " end"))))
               (pe-call oper args ctx))))

        ((fut? ptree)
         (error "future special form not supported"))

        (else
         (error "unknown parse tree node type" ptree))))

(define (indent n)
  (vector->list (make-vector n "  ")))

(define (pe-constant val)
  (cond ((boolean? val)
         (if val "true" "false"))
        ((number? val)
         (number->string
          (if (and (exact? val) (integer? val))
              val
              (exact->inexact val))))
        ((char? val)
         (let ((n (char->integer val)))
           (if (and (>= n 33) (<= n 126))
               (string #\$ val)
               n)))
        ((string? val)
         (object->string val))
        ((symbol? val)
         (let* ((str (symbol->string val))
                (lst (string->list str)))
           (if (and (pair? lst)
                    (char>=? (car lst) #\a)
                    (char<=? (car lst) #\z)
                    (not (memq #f (map (lambda (c)
                                         (or (char-alphabetic? c)
                                             (char=? c #\_)
                                             (char=? c #\@)))
                                       lst))))
               str
               (list "'" str "'"))))
        ((null? val)
         "[]")
        ((pair? val)
         (pe-list val "[" "]" pe-constant))
        ((vector? val)
         (pe-list (vector->list val) "{" "}" pe-constant))
        ((u8vector? val)
         (pe-list (u8vector->list val) "<<" ">>" pe-constant))
        (else
         (error "Scheme constant cannot be converted to Erlang" val))))

(define (pe-list lst open close cvt)
  (if (pair? lst)
      (let loop ((lst (cdr lst))
                 (rev-res (list (cvt (car lst)) open)))
        (if (pair? lst)
            (loop (cdr lst)
                  (cons (cvt (car lst)) (cons "," rev-res)))
            (reverse
             (if (null? lst)
                 (cons close rev-res)
                 (cons close (cons (cvt lst) (cons "|" rev-res)))))))
      (string-append open close)))

(define (pe-call oper args ctx)
  (let ((a (map (lambda (x) (pe x (ctx-body?-set ctx #f))) args)))
    (cond ((and (ref? oper)
                (not (assq (ref-var oper) ;; top-level function?
                           (vector-ref (ctx-lvars ctx) 0))))
           (let ((name (var-name (ref-var oper))))
             (or (pe-builtin-call name a)
                 (list (pe-fun-name name)
                       (pe-list a "(" ")" (lambda (x) x))))))
          ((and (cst? oper)
                (proc-obj? (cst-val oper)))
           (let ((name (string->symbol (proc-obj-name (cst-val oper)))))
             (or (pe-builtin-call name a)
                 (error "Unimplemented primitive:" name))))
          (else
           (list (pe oper (ctx-body?-set ctx #f))
                 (pe-list a "(" ")" (lambda (x) x)))))))

(define (pe-builtin-call name a)
  (let ((x (assq name prim-db)))
    (and x
         (let ((descr (cdr x)))
           (if (if (< (car descr) 0)
                   (>= (length a) (- -1 (car descr)))
                   (= (length a) (car descr)))
               (apply (cdr descr) a)
               (error "wrong number of arguments; primitive:" name))))))

(define (foldr-bin-op* op a)
  (let loop ((lst (cdr a))
             (rev-res (list (car a))))
    (if (null? lst)
        (cons "(" (reverse (cons ")" rev-res)))
        (loop (cdr lst)
              (cons (car lst) (cons " " (cons op (cons " " rev-res))))))))

(define (foldr-bin-op op a base)
  (if (null? a)
      (pe-constant base)
      (foldr-bin-op* op a)))

(define prim-db '())

(define (def-prim name descr)
  (set! prim-db (cons (cons name descr) prim-db)))

(define (def-fixed name nargs generator)
  (def-prim name (cons nargs generator)))

(define (def-nary name nargs generator)
  (def-prim name (cons (- -1 nargs) generator)))

(def-fixed 'apply         2 (lambda (x y) (list "apply(" x "," y ")")))
(def-fixed 'not           1 (lambda (x) (list "(not " x ")")))
(def-nary  '+             0 (lambda args (foldr-bin-op "+" args 0)))
(def-nary  '-             1 (lambda args (foldr-bin-op* "-" args)))
(def-nary  '*             0 (lambda args (foldr-bin-op "*" args 1)))
(def-nary  '/             1 (lambda args (foldr-bin-op* "/" args)))
(def-fixed 'quotient      2 (lambda (x y) (list "(" x " div " y ")")))
(def-fixed 'remainder     2 (lambda (x y) (list "(" x " rem " y ")")))
(def-fixed '=             2 (lambda (x y) (list "(" x " == " y ")")))
(def-fixed '<             2 (lambda (x y) (list "(" x " < " y ")")))
(def-fixed '>             2 (lambda (x y) (list "(" x " > " y ")")))
(def-fixed '<=            2 (lambda (x y) (list "(" x " =< " y ")")))
(def-fixed '>=            2 (lambda (x y) (list "(" x " >= " y ")")))
(def-fixed 'abs           1 (lambda (x) (list "abs(" x ")")))
(def-fixed 'truncate      1 (lambda (x) (list "trunc(" x ")")))
(def-fixed 'round         1 (lambda (x) (list "round(" x ")")))
(def-fixed 'cons          2 (lambda (x y) (list "[" x "|" y "]")))
(def-fixed 'car           1 (lambda (x) (list "hd(" x ")")))
(def-fixed 'cdr           1 (lambda (x) (list "tl(" x ")")))
(def-nary  'list          0 (lambda args (pe-list args "[" "]" (lambda (x) x))))
(def-fixed 'length        1 (lambda (x) (list "length(" x ")")))
(def-nary  'append        0 (lambda args (foldr-bin-op "++" args '())))
(def-fixed 'null?         1 (lambda (x) (list "(" x " == [])")))
(def-fixed 'map           2 (lambda (x y) (list "lists:map(" x "," y ")")))
(def-fixed 'procedure?    1 (lambda (x) (list "is_function(" x ")")))
(def-fixed 'symbol?       1 (lambda (x) (list "is_atom(" x ")")))
(def-fixed 'symbol->string 1 (lambda (x) (list "atom_to_list(" x ")")))
(def-fixed 'string->symbol 1 (lambda (x) (list "list_to_atom(" x ")")))
(def-fixed 'number?       1 (lambda (x) (list "is_number(" x ")")))
(def-fixed 'boolean?      1 (lambda (x) (list "is_boolean(" x ")")))
(def-fixed 'u8vector?     1 (lambda (x) (list "is_binary(" x ")")))
(def-fixed 'u8vector-length 1 (lambda (x) (list "size(" x ")")))
(def-fixed 'u8vector->list 1 (lambda (x) (list "binary_to_list(" x ")")))
(def-fixed 'list->u8vector 1 (lambda (x) (list "list_to_binary(" x ")")))
(def-fixed 'vector?       1 (lambda (x) (list "is_tuple(" x ")")))
(def-fixed 'vector-length 1 (lambda (x) (list "size(" x ")")))
(def-fixed 'vector-ref    2 (lambda (x y) (list "element(" y "+1," x ")")))
(def-fixed 'vector->list  1 (lambda (x) (list "tuple_to_list(" x ")")))
(def-fixed 'list->vector  1 (lambda (x) (list "list_to_tuple(" x ")")))
(def-nary  'make-vector   1
  (lambda args
    (case (length args)
      ((1)
       (list "erlang:make_tuple(" (list-ref args 0) ",false)"))
      ((2)
       (list "erlang:make_tuple(" (list-ref args 0) "," (list-ref args 1) ")"))
      (else
       (error "wrong number of arguments; primitive:" 'make-vector)))))
(def-fixed '##box         1 (lambda (x) (list "schemelib:box(" x ")")))
(def-fixed '##unbox       1 (lambda (x) (list "schemelib:unbox(" x ")")))
(def-fixed '##set-box!    2 (lambda (x y) (list "schemelib:setbox(" x "," y ")")))
(def-fixed 'display       1 (lambda (x) (list "erlang:display(" x ")")))
(def-fixed 'spawn         1 (lambda (x) (list "spawn(" x ")")))
(def-fixed 'remote-spawn  2 (lambda (x y) (list "spawn(" x "," y ")")))
(def-fixed 'self          0 (lambda () (list "self()")))
(def-fixed '!             2 (lambda (x y) (list "(" x " ! " y ")")))
(def-nary  '?             0
  (lambda args
    (let ((v (newtmp)))
      (case (length args)
        ((0)
         (list "receive " v " -> " v " end"))
        ((1)
         (list "receive " v " -> " v " after round(" (list-ref args 0) "*1000) -> throw timeout end"))
        ((2)
         (list "receive " v " -> " v " after round(" (list-ref args 0) "*1000) -> " (list-ref args 1) " end"))
        (else
         (error "wrong number of arguments; primitive:" '?))))))

(let ((args (command-line)))
  (for-each s2e (cdr args)))

;;;============================================================================
